#include "graph.h"

void write_signal(const valarray<complex<double>> &s, const vector<double> &sx, const string filename)
{
    ofstream file(filename.c_str());
    file << "#t real imag" << endl;
    for (size_t i = 0; i < s.size(); i++)
    {
        file << sx[i] << ' ' << s[i].real() << ' ' << s[i].imag() << endl;
    }
    file.close();
}

void write_data_plot(const vector<vector<double>> &data, const vector<double> &x_axis,
                     string filename)
{
    ofstream file(filename.c_str());

    int Mr = data.size();
    int Nr = data[0].size();
    for (int i = 0; i < Mr; i++)
    {
        file << x_axis[i] << ' ';
        for (int j = 0; j < Nr; j++)
        {
            file << data[i][j] << ' ';
        }
        file << endl;
    }
    file.close();
}

void write_3d_data_plot(const vector<vector<double>> &data,
                        const vector<double> &x_axis,
                        string filename, bool columndata, bool doppler=false)
{
    ofstream file(filename.c_str());

    int Mr = data.size();
    int Nr = data[0].size();
    if (columndata)
    {
        for (int i = 0; i < Nr; i++)
        {
            for (int j = 0; j < Mr; j++)
            {
                file << ((doppler) ? double(i) / 100 : i) << ' ' << x_axis[j] << ' ' << data[j][i] << '\n';
            }
            file << endl;
        }
    }
    else
    {
        for (int i = 0; i < Mr; i++)
        {
            for (int j = 0; j < Nr; j++)
            {
                file << ((doppler) ? double(i) / 100 : i) << ' ' << x_axis[j] << ' ' << data[i][j] << '\n';
            }
            file << endl;
        }
    }
    file.close();
}

void doppler_process(const vector<valarray<complex<double>>> &y,
                     vector<valarray<complex<double>>> &yp,
                     vector<double> &w, int Lfft)
{
    //	vector<valarray<complex<double> > > yp;
    int My = y.size();
    int Ny = y[0].size();
    int N = (Lfft <= Ny) ? Ny : Lfft;
    for (int i = 0; i < My; i++)
    {
        valarray<complex<double>> in(N);
        vector<double> YdBi;

        YdBi.clear();
        for (int j = 0; j < Ny; j++)
        {
            in[j] = conj(y[i][j]) * w[j];
        }

        fft(in, IsPowerofTwo(N));
        in = in * in.apply(std::conj);
        yp.push_back(in);
    }
    git_rotate(yp, Lfft / 2);
}

void range_doppler_plot(const vector<valarray<complex<double>>> &y,
                        const vector<double> &range, int Lfft, string filename)
{
    //	int My = y.size();
    int Ny = y[0].size();
    vector<double> w = hamming(Ny);

    vector<valarray<complex<double>>> yp;
    vector<vector<double>> YdB;
    doppler_process(y, yp, w, Lfft);
    int M = yp.size();
    int N = yp[0].size();
    cout << yp.size() << " " << yp[0].size() << endl;

    double maxdb = 0.0;

    for (int i = 0; i < M; i++)
    {
        vector<double> YdBi;

        for (int j = 0; j < N; j++)
        {
            double val = db(yp[i][j], "power");
            if (val > maxdb)
                maxdb = val;
            YdBi.push_back(val);
        }
        YdB.push_back(YdBi);
    }

    cout << filename << ":  " << maxdb << endl;
    write_3d_data_plot(YdB, range, filename, false, true);
}

void range_plot(vector<valarray<complex<double>>> &y,
                const vector<double> &range, int Lfft, string filename)
{
    int M = y.size();
    int N = y[0].size();
    cout << y.size() << " " << y[0].size() << endl;
    vector<vector<double>> YdB;
    double maxdb = 0.0;

    for (int i = 0; i < M; i++)
    {
        vector<double> YdBi;

        for (int j = 0; j < N; j++)
        {
            double val = db(y[i][j], "power");
            if (val > maxdb)
                maxdb = val;
            YdBi.push_back(val);
        }
        YdB.push_back(YdBi);
    }

    cout << filename << ":  " << maxdb << endl;
    write_3d_data_plot(YdB, range, filename, true);
}

void simulate_signal_plot(const radar &a_radar, const vector<valarray<complex<double>>> &y)
{
    /* output the data for plots */
    vector<double> x_axis;
    for (size_t i = 0; i < y.size(); i++)
    {
        double val = ((double)C / 2.0) * ((i * (1.0 / a_radar.chirp_sampling_rate)) + a_radar.pulse_start_times[0]) / (double)1e3;
        x_axis.push_back(val);
    }

    int My = y.size();
    int Ny = y[0].size();
    vector<vector<double>> ydB, ydBscaled;
    double ydBmax = 0.0;
    for (int i = 0; i < My; i++)
    {
        vector<double> ydbi;
        vector<double> ydbscaledi;
        ydbi.clear();
        ydbscaledi.clear();
        for (int j = 0; j < Ny; j++)
        {
            ydbi.push_back(db(y[i][j], "voltage"));
            double val = abs(y[i][j]);
            if (val > ydBmax)
                ydBmax = val;
            ydbscaledi.push_back(val);
        }
        ydB.push_back(ydbi);
        ydBscaled.push_back(ydbscaledi);
    }
    write_data_plot(ydB, x_axis, "y.dat");

    for (int i = 0; i < My; i++)
    {
        for (int j = 0; j < Ny; j++)
        {
            ydBscaled[i][j] = db(ydBscaled[i][j] / ydBmax, "voltage");
        }
    }
    write_3d_data_plot(ydBscaled, x_axis, "ydb.dat", true);
}