#include "radar.h"
#include "inc.h"
#include "util.h"

extern double PRF;

bool Target::operator<(Target &rhs)
{
    if (range == rhs.range)
    {
        return velocity < rhs.velocity;
    }

    return range < rhs.range;
}

valarray<complex<double>> chirp_signal(double T, double W, double p)
{

	double delta_t = 1.0 / (p * W); // 1 / 12e6
	int N = p * T * W;				// 120
	valarray<complex<double>> x(N);
	for (int i = 0; i < N; i++)
	{
		double d = pow((delta_t * i - (double)(N - 1) / 2.0 / p / W), 2);
		double j = M_PI * W / T * d;
		complex<double> t(0, j);
		x[i] = exp(t);
	}

	return x;
}

/*
 * Print out the radar information to the given outfile.
 */
void radar::print_to_file(ostream &outfile)
{
	outfile << "PRF: " << PRF << endl;
	outfile << "PL: " << pulse_length << endl;
	outfile << "CB: " << chirp_bandwidth << endl;
	outfile << "CSR: " << chirp_sampling_rate << endl;
	outfile << "PN: " << pulse_number << endl;
	outfile << "STWR: " << start_and_end[0] << endl;
	outfile << "ETWR: " << start_and_end[1] << endl;
	outfile << "RF: " << fc << endl;
	outfile << "NT: " << targets.size() << endl;
	for (Target &target : targets)
	{
		outfile << target.range << " " << target.SNR << " " << target.velocity << endl;
	}
}

/*
 * Initialize the radar class.
 */
void radar::setup(istream& infile)
{
	string junk;
	infile >> junk >> PRF;

	/* form radar chirp pulse */
	infile >> junk >> pulse_length;		// pulse length, sec
	infile >> junk >> chirp_bandwidth;		// chirp bandwidth, Hz
	infile >> junk >> chirp_sampling_rate; // chirp sampling rate, Hz

	// Create chirp signal that will be used to simulate signals from pulse doppler radar
	chirp = chirp_signal(pulse_length, chirp_bandwidth, chirp_sampling_rate / chirp_bandwidth);

	infile >> junk >> pulse_number; // Number of pulses
	double PRI = 1.0 / PRF;		 // PRI in sec

	for (int i = 0; i < pulse_number; i++)
	{
		pulse_start_times.push_back(PRI * (double)i);
		gains.push_back(1.0);
	}
	start_and_end = vector<double>(2); // start and end times of range window in sec
	infile >> junk >> start_and_end[0];
	infile >> junk >> start_and_end[1];

	infile >> junk >> fc;

	/* compute unambiguous Doppler interval in m/sec
	 * compute unambiguous range interval in meters
	 */
	double vua = C * PRF / (2 * fc);
	double rmin = C * start_and_end[0] / 2;
	double rmax = C * start_and_end[1] / 2;
	double rua = C / 2 / PRF;

	cout << "the unambiguous velocity interval is " << vua << "m/s" << endl;
	cout << "the range window starts at " << rmin / 1e3 << "km" << endl;
	cout << "the range window ends at " << rmax / 1e3 << "km" << endl;
	cout << "the unambiguous range interval is " << rua / 1e3 << "km" << endl;

	/* Read in the targets */
	int num_targets;
	infile >> junk >> num_targets;

	double range, snr, vel;
	while (num_targets-- && infile >> range >> snr >> vel)
	{
		targets.push_back(Target{range, vel, snr});
	}

	/* Sort to for easier lookup later */
	sort(targets.begin(), targets.end());

	cout << "\nInputted Targets:\n";
	cout << "Target #\tRange (m)\tSNR\tVelocity (m/s)\n";
	for (size_t i = 0; i < targets.size(); ++i)
	{
		cout << i << "\t\t" << targets[i].range << "\t\t" << targets[i].SNR << "\t" << targets[i].velocity << "\n";
	}
	cout << endl;

	/* 
	 * Compute relative RCS using the idea that SNR is proportional to
	 * RCS/R^4
	 */
	vector<double> rel_RCS; // Radar cross section (determines how detectable an object is)
	double max_rcs = 0.0;
	for (size_t i = 0; i < targets.size(); i++)
	{
		double val = pow(10, targets[i].SNR / 10) * pow(targets[i].range, 4);
		if (val > max_rcs)
		{
			max_rcs = val;
		}
		rel_RCS.push_back(val);
	}

	// convert to power
	for (size_t i = 0; i < rel_RCS.size(); i++)
	{
		double temp = rel_RCS[i];
		rel_RCS[i] = db(temp / max_rcs, "power");
	}
}

/*
 * Generate the signal matrix that contains if all signals.
 * Adds noise to mimic a standard signal and changes the signal according
 * to the SNR as well.
 */
vector<valarray<complex<double>>> radar::simulate()
{
	chrono::time_point<std::chrono::system_clock> start_time, end_time; // Start and end times
	start_time = chrono::system_clock::now();
	vector<valarray<complex<double>>> y;
	int chirp_size = chirp.size();
	double sample_interval = 1.0 / chirp_sampling_rate; // sampling interval (sec)

	//output sampling times (sec) from start to end with step sample_interval
	vector<double> sample_range = sampling(pulse_start_times[0], sample_interval, pulse_start_times[1]);
	double input_pulse_length = chirp_size * sample_interval; // length of input pulse (sec)

	/* 
	 * Determine the quadratic phase modulation parameters for later
	 * interpolation of pulse samples
	 */
	vector<double> t_x;
	for (int i = 0; i < chirp_size; i++)
	{
		t_x.push_back(i * sample_interval * 1e9); // use msec for quadratic fitting
	}
	vector<double> x_ph;
	for (int i = 0; i < chirp_size; i++)
	{
		x_ph.push_back(arg(chirp[i]));
	}
	unwrap(x_ph);
	vector<double> q; // coefficients of the quadratic function
	quadraticfit(t_x, x_ph, q);

	int Mr = sample_range.size(); //rows of output matrix y
	int Nr = gains.size();		  // cols of output matrix y
	initVector(Mr, Nr, y);

	/*'i' loops over the number of targets*/
	for (size_t i = 0; i < targets.size(); i++)
	{
		double ri = targets[i].range;
		double vi = targets[i].velocity;
		double f_doppler = 2.0 * vi * fc / C;

		/*'j' loops over the number of pulses*/
		for (size_t j = 0; j < gains.size(); j++)
		{
			double r_at_T_0 = ri - vi * pulse_start_times[j];

			/**
			 * compute start and end time of reflected pulse at receiver
			 * ensure that it falls at least partially within the range window
			 */
			double tau = 2.0 * r_at_T_0 / (C + vi);
			double tmax = tau + input_pulse_length;

			if (tau >= pulse_start_times[1] || tmax <= pulse_start_times[0])
			{
				cout << "Echo from target " << i << " at range " << ri << "km" << endl;
				cout << "is COMPLETELY OUT OF the range window" << endl;
				cout << "on pulse " << j << endl;
			}
			else
			{
				/* place scaled, range-delayed, Doppler shifted pulse
				 * into output matrix
				 * unit noise power and unit nominal pulse amplitude assumed
				 * to get amplitude from SNR*/
				double amp = pow(10, targets[i].SNR / 20.0);
				for (size_t t = 0; t < sample_range.size(); t++)
				{
					/*
					 * figure out which sample locations in the output grid contain
					 * reflected pulse
					 * */
					double t_vals_t = sample_range[t] - tau;
					if (t_vals_t >= 0 && t_vals_t < input_pulse_length)
					{
						complex<double> t1(0, -2.0 * M_PI * fc * tau);
						complex<double> t2(0, 2.0 * M_PI * f_doppler * sample_range[t]);
						// convert t_vals_t to msec for polyval
						complex<double> t3(0, polyval(q, t_vals_t * 1e9));
						//						cout << exp(t1) << " " << exp(t2) << " " << exp(t3) << endl;
						y[t][j] = y[t][j] + (amp * gains[j] * exp(t1)) * exp(t2) * exp(t3);
					}
				}
			}
		}
	}
	add_noise(y);
	end_time = chrono::system_clock::now();
	cout << "\nSignal simulation time: " << chrono::duration<double>(end_time - start_time).count() << "s\n\n";
	return y;
}

/*
 * Add noise to the signal matrix by randomly increasing and decreasing values
 * over a normal distribution.
 */
void radar::add_noise(vector<valarray<complex<double>>> &y)
{
	int My = y.size();
	int Ny = y[0].size();

	vector<vector<double>> real(My, vector<double>(Ny));
	vector<vector<double>> imag(My, vector<double>(Ny));
	vector<vector<double>> r(My, vector<double>(Ny));

#if RANDOM
	std::default_random_engine generator((unsigned)time(NULL) * getpid());
#else
	std::default_random_engine generator(0);
#endif
	std::normal_distribution<double> distribution(0, 1);

	for (int i = 0; i < My; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			real[i][j] = distribution(generator);
		}
	}

	for (int i = 0; i < My; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			imag[i][j] = distribution(generator);
		}
	}

	for (int i = 0; i < My; ++i)
	{
		for (int j = 0; j < Ny; ++j)
		{
			r[i][j] = distribution(generator);
		}
	}

	/* adding noise */
	for (int i = 0; i < My; i++)
	{
		for (int j = 0; j < Ny; j++)
		{
			complex<double> white_noise(real[i][j], imag[i][j]);
			y[i][j] += y[i][j] + (1.0 / (double)sqrt(2.0)) * white_noise;
		}
	}

	/* creating clutter
	 * create log-normal (ground) "clutter" with specified C/N
	 * and log-normal standard deviation for amplitude, uniform phase
	 * clutter is uncorrelated in range, fully correlated in pulse
	 */
	int CN = 20;   // clutter-to-noise ratio in first bin (dB)
	int SDxdB = 3; // in dB (is not the sigma of the complete clutter)
	vector<valarray<complex<double>>> ncc;
	for (int i = 0; i < My; i++)
	{
		valarray<complex<double>> ncci(Ny);
		for (int j = 0; j < Ny; j++)
		{
			double val = pow(10.0, SDxdB * r[i][j] / 10.0);
			complex<double> v1(0, 2.0 * M_PI * rand_uniform());
			ncci[j] = val * exp(v1);
		}
		ncc.push_back(ncci);
	}

	/* correlating and adding clutter
	 * force the power spectrum shape to be Gaussian*/
	vector<double> G;
	double gau[5];
	for (int i = 0; i < 5; i++)
	{
		gau[i] = exp(-(i * i) / 1.0);
	}

	for (int i = 0; i < Ny; i++)
	{
		if (i < 5)
			G.push_back(gau[i]);
		else if (i >= Ny - 5 + 1)
		{
			G.push_back(gau[Ny - i]);
		}
		else
			G.push_back(0.0);
	}

	complex<double> mean(0.0, 0.0);
	complex<double> var(0.0, 0.0);
	for (int i = 0; i < My; i++)
	{
		fft(ncc[i], IsPowerofTwo(Ny));

		for (int j = 0; j < Ny; j++)
		{
			ncc[i][j] = ncc[i][j] * G[j];
		}

		ifft(ncc[i], IsPowerofTwo(Ny));
		mean += ncc[i].sum();
	}
	mean = mean / (double)(My * Ny);
	//	cout << mean << endl;
	for (int i = 0; i < My; i++)
	{
		for (int j = 0; j < Ny; j++)
		{
			complex<double> nc = ncc[i][j];
			var += (nc - mean) * conj(nc - mean);
		}
	}

	var = var / (double)(My * Ny);

	/* rescale clutter to have desired C/N ratio */
	double scale = sqrt((double)pow(10, CN / 10) / var.real());
	//	cout << scale << endl;
	for (int i = 0; i < My; i++)
	{
		/* weight the clutter power in range for assume R^2 (beam-limited) loss */
		double cweight = pulse_start_times[0] * pow(pulse_start_times[0] + ((i == 0) ? 10e-10 : i) * (1.0 / chirp_sampling_rate), -1);
		for (int j = 0; j < Ny; j++)
		{
			complex<double> nc = ncc[i][j];
			y[i][j] = y[i][j] + cweight * scale * nc;
		}
	}
}
