/*
 * simulator.cpp
 *
 *  Created on: 21/05/2014
 *      Author: yan
 */

#ifndef SIMULATOR_CPP_
#define SIMULATOR_CPP_
#include "graph.h"
#include "inc.h"
#include "util.h"
#include "radar.h"


pair<double, double> error_percentage(const Target& actual, const Target& estimated) {
	double range_p = abs(abs(actual.range - estimated.range) / actual.range * 100);
	double vel_p = abs(abs(actual.velocity - estimated.velocity) / actual.velocity * 100);
	return {range_p, vel_p};
}


/*
 * Initialize a radar with the passed in parameters.
 * 
 * Parameters from stdin:
 * 
 * pulse_repition_frequency: Number of pulses in a second (pulses / second), default 10e3
 * pulse_length: How long the pulse is sent for (sec), default 10e-6
 * chirp_bandwidth: Bandwidth for the chirp signal that simulates a doppler signal (Hz), default 10e6
 * chirp_sampling_rate: How often the chirp signal is sampled (Hz), default 12e6
 * pulse_number: Number of pulses sent in the simulation, default 20
 * {start/end}_time_range_window: Beginning and end of distance an object can be (sec), default 12e-6 (start) and 40e-6 (end)
 * rf_frequency: radio frequency (Hz), default 10e9
 * targets: List of targets given by distance (m), SNR, velocity (m/s)
 */
void get_a_radar(radar &a_radar, istream& infile)
{
	// Initialize radar with created values
	a_radar.setup(infile);
}

/* Get the estimated error between the actual target and the estimated. */
void error_calculations(const vector<Target> &actual, const vector<Target> &estimated)
{
	cout << "\nError Percentages:" << endl;
	cout << "Target #\tRange (%)\tVelocity (%)" << endl;
	cout << fixed << setprecision(2);
	for (size_t i = 0; i < actual.size(); ++i)
	{
		if (estimated[i].range < 0) {
			cout << i << "\t\tN/A\t\tN/A" << endl;
		}
		else {
			pair<double, double> error = error_percentage(actual[i], estimated[i]);

			cout << i << "\t\t" << error.first << "\t\t" << error.second << endl;
		}
	}
}

// Iteratively attempt to find the lowest error targets according to range
vector<int> get_lowest_range_errors(const Target& target, const vector<Target>& search, unordered_set<int>& used_index, double& max_error) {
	vector<int> in_range_targets;
	double lowest_error = 0.0;
	max_error = 0.0;

	// Try to get all within range error percentage of 5%, then 10%, 15%, etc.
	while(in_range_targets.empty()) {
		max_error = lowest_error + 5.0;
		lowest_error = numeric_limits<double>::max();

		for (int i = 0; i < int(search.size()); ++i) {
			if (!used_index.count(i)) {
				double error = error_percentage(target, search[i]).first;
				if (!used_index.count(i) && error < max_error) {
					in_range_targets.push_back(i);
				}
				else if (error < lowest_error) {
					lowest_error = error;
				}
			}
		}
	}

	return in_range_targets;
}

/* Find the matching targets for each given target */
vector<Target> find_matching_targets(const vector<Target>& correct, const vector<Target>& detected) {
	vector<Target> matching;
	unordered_set<int> used_index;
	matching.reserve(correct.size());
	if (correct.size() <= detected.size()) {
		for (const Target& target : correct) {
			// Get closest according to range
			double max_error;
			vector<int> in_range_targets = get_lowest_range_errors(target, detected, used_index, max_error);
			
			// Out of the remaining, find the best one in terms of velocity
			int best_index = in_range_targets[0];
			double lowest_error = max_error + 1.0;
			for (int i = 0; i < int(in_range_targets.size()); ++i) {
				if (error_percentage(target, detected[in_range_targets[i]]).second < lowest_error) {
					best_index = in_range_targets[i];
					lowest_error = error_percentage(target, detected[in_range_targets[i]]).second;
				}
			}

			// Add the best one and remove from detected
			matching.push_back(detected[best_index]);
			used_index.insert(best_index);
		}
	}
	else {
		matching.resize(correct.size());

		for (const Target& target : detected) {
			// Try to get all within range error percentage of 5%, then 10%, 15%, etc.
			double max_error = 0.0;
			vector<int> in_range_targets = get_lowest_range_errors(target, correct, used_index, max_error);

			// Out of the remaining, find the best one in terms of velocity
			int best_index = in_range_targets[0];
			double lowest_error = max_error + 1.0;
			for (int i = 0; i < int(in_range_targets.size()); ++i) {
				if (error_percentage(target, correct[in_range_targets[i]]).second < lowest_error) {
					best_index = in_range_targets[i];
					lowest_error = error_percentage(target, correct[in_range_targets[i]]).second;
				}
			}

			matching[best_index] = target;
			used_index.insert(best_index);
		}
	}
	return matching;
}

/*
 * Using the signals in the signal matrix, y, calculate the distances using doppler calculation
 * 
 * Steps
 * 1. Time-domain hamming weighting of the impulse response for range sidelobe control.
 * 2. Apply 3-pulse canceller to remove clutter. Create better quality signal
 * 3. Search for the range bins with targets first noncoherently integrate across the frequency bins
 * 4. find the Doppler peak(s) for each range bin having a target(s). 
 *    Keep adjoining Doppler values as well to support subsequent interpolation
 * 
 */
void process_radar_signal(radar &a_radar, vector<valarray<complex<double>>> &y)
{
	vector<double> sx;
	for (size_t i = 0; i < a_radar.chirp.size(); i++)
	{
		double val = ((double)1e6 / a_radar.chirp_sampling_rate) * i;
		sx.push_back(val);
	}
#ifdef GENPLOT
	write_signal(a_radar.chirp, sx, "chirp.dat");
#endif

	//compute unambiguous Doppler interval in m/sec
	double vua = (-1) * C * a_radar.PRF / (2 * a_radar.fc);

	//convert range samples to absolute range units
	int My = y.size();
	int Ny = y[0].size();
	vector<double> range;
	for (int i = 0; i < My; i++)
	{
		double val = (double(C) / 2.0) * ((double)i * (1.0 / a_radar.chirp_sampling_rate) + a_radar.pulse_start_times[0]) / 1e3;
		range.push_back(val);
	}
	vector<int> pulses;
	for (int i = 0; i < Ny; i++)
	{
		pulses.push_back(i);
	}

	//force oversize FFT, and compute doppler scale factor
	int p = nextpow2(Ny);
	int Lfft = pow(2.0, p + 3);
	vector<double> doppler;
	for (int i = 0; i < Lfft; i++)
	{
		double val = ((double)i / (double)Lfft - 0.5) * vua;
		doppler.push_back(val);
	}

	/* doppler process and square-law detect the whole
	 * unprocessed array
	 * using Hamming window throughout*/
#ifdef GENPLOT
	range_doppler_plot(y, doppler, Lfft, "doppler_range.dat");
#endif
	chrono::time_point<std::chrono::high_resolution_clock> start_time, end_time; // Start and end times
	double total_time = 0;
	/* processing the data
	 * 1. pulse compression
	 * 	  using time-domain hamming weighting of the
	 * 	  impulse response for range sidelobe control */
	start_time = chrono::high_resolution_clock::now();
	int Ls = a_radar.chirp.size();
	vector<double> hw = hamming(Ls);
	valarray<complex<double>> h1(Ls);
	for (int i = 0; i < Ls; i++)
	{
		h1[i] = conj(a_radar.chirp[Ls - 1 - i]) * hw[i];
	}
	int Myp = My + Ls - 1;
	int Nyp = Ny;
	vector<valarray<complex<double>>> yp;
	initVector(Myp, Nyp, yp);
	conv(h1, y, yp, false);
	vector<double> rangep;
	for (int i = 0; i < Myp; i++)
	{
		double rval = (C / 2.0) * ((double(i) - (Ls - 1)) * (1.0 / a_radar.chirp_sampling_rate) + a_radar.pulse_start_times[0]) / 1e3;
		rangep.push_back(rval);
	}
	end_time = chrono::high_resolution_clock::now();
	cout << "\nPulse Compression Duration: " << chrono::duration<double>(end_time - start_time).count() << "s\n";
	total_time += chrono::duration<double>(end_time - start_time).count();
#ifdef GENPLOT
	range_doppler_plot(yp, doppler, Lfft, "doppler_range_compressed.dat");
#endif

	/* 2. apply three-pulse canceller in each range bin to raw data */
	start_time = chrono::high_resolution_clock::now();
	valarray<complex<double>> h2(3);
	h2[0] = complex<double>(1, 0);
	h2[1] = complex<double>(-2, 0);
	h2[2] = complex<double>(1, 0);
	vector<valarray<complex<double>>> ypm;
	int Nyp2 = Nyp + h2.size() - 1;
	initVector(Myp, Nyp2, ypm);
	conv(h2, yp, ypm, true);
	vector<valarray<complex<double>>> YPM;
	vector<double> w = hamming(Ny);
	doppler_process(ypm, YPM, w, Lfft);
	end_time = chrono::high_resolution_clock::now();
	cout << "3-Pulse Canceller Duration: " << chrono::duration<double>(end_time - start_time).count() << "s\n";
	total_time += chrono::duration<double>(end_time - start_time).count();
#ifdef GENPLOT
	range_doppler_plot(ypm, doppler, Lfft, "doppler_range_canceller.dat");
#endif

	/*3. Search for the range bins with targets
	 *   first noncoherently integrate across the frequency bins*/
	start_time = chrono::high_resolution_clock::now();
	int Mypm = YPM.size();
	//	int Nypm = YPM[0].size();
	valarray<double> YPMrange(Mypm);
	vector<double> ranges_sort;
	for (int i = 0; i < Mypm; i++)
	{
		double sumr = YPM[i].sum().real();
		YPMrange[i] = sumr;
		ranges_sort.push_back(sumr);
	}
	sort(ranges_sort.begin(), ranges_sort.end());
	/* median of YPMrange */
	int md_ind = (Mypm % 2) ? Mypm / 2 : (Mypm / 2 + 1);
	double Nrange = ranges_sort[md_ind];
	/* threshold 8x (9DB) above noise estimate */
	double Trange = 8 * Nrange;

	/* loop identifies which range bins have local peaks above
	 * the threshold. It also set up a vector */
	vector<pair<int, double>> spikesr;
	for (int i = 2; i < Mypm - 1; i++)
	{
		if ((YPMrange[i] > YPMrange[i + 1]) && (YPMrange[i] > YPMrange[i - 1]) && (YPMrange[i] > Trange))
		{
			spikesr.push_back(make_pair(i, YPMrange[i]));
		}
	}

	end_time = chrono::high_resolution_clock::now();
	cout << "Range Bins Search Duration: " << chrono::duration<double>(end_time - start_time).count() << "s\n";
	total_time += chrono::duration<double>(end_time - start_time).count();

	vector<Target> targets;
	/*  4. find the Doppler peak(s) for each range bin having a target(s).  Keep
		adjoining Doppler values as well to support subsequent interpolation
	*/
	start_time = chrono::high_resolution_clock::now();
	int Mspikesr = spikesr.size();
	for (int i = 0; i < Mspikesr; i++)
	{
		vector<vector<double>> spikesd_bin;
		spikesd_bin.clear();
		int rb = spikesr[i].first;
		for (int k = 0; k < Lfft; k++)
		{
			int km1 = k == 0 ? Lfft - 1 : k - 1;
			int kp1 = k == Lfft - 1 ? 0 : k + 1;

			if (YPM[rb][k].real() > YPM[rb][kp1].real() &&
				YPM[rb][k].real() > YPM[rb][km1].real() &&
				YPM[rb][k].real() > (double)Nrange / 2.0)
			{
				vector<double> spikesd;
				vector<double> obj;
				spikesd.clear();
				obj.clear();
				//				cout << k << " " <<YPM[rb][km1].real()<<" "<<YPM[rb][k].real()
				//						<< " " << YPM[rb][kp1].real() << endl;
				spikesd.push_back(k);
				spikesd.push_back(YPM[rb][km1].real());
				spikesd.push_back(YPM[rb][k].real());
				spikesd.push_back(YPM[rb][kp1].real());
				double amp, del_k;
				double z1 = sqrt(spikesd[1]);
				double z2 = sqrt(spikesd[2]);
				double z3 = sqrt(spikesd[3]);
				peakinterp(amp, del_k, z1, z2, z3);
				obj.push_back(amp);
				obj.push_back(rangep[rb]);

				double vel = ((spikesd[0] + del_k - 1) / Lfft - 0.5) * vua;
				obj.push_back(vel);

				spikesd_bin.push_back(spikesd);
				targets.push_back(Target{obj[1] * 1000, obj[2], 0.0});
			}
		}
	}
	end_time = chrono::high_resolution_clock::now();
	cout << "Doppler Peak Search: " << chrono::duration<double>(end_time - start_time).count() << "s\n";
	total_time += chrono::duration<double>(end_time - start_time).count();
	cout << "Total doppler process time: " << total_time << "s\n\n";

	sort(targets.begin(), targets.end());
	if (targets.empty())
	{
		cout << "No targets detected." << endl;
	}
	else
	{
		cout << "Detected targets:" << endl;
		cout << "Range (m)\tVelocity (m/s)" << endl;
		int Mtargets = targets.size();
		for (int i = 0; i < Mtargets; i++)
		{
			cout << targets[i].range << "\t\t" << targets[i].velocity << "\n";
		}
	}

	if (a_radar.targets.size() != targets.size())
	{
		cout << "\nERROR: There were " << a_radar.targets.size() << " targets, but ";
		cout << targets.size() << " were detected.\n";
		cout << "Attempting to match with closest detected.\n";

		vector<Target> matching_target = find_matching_targets(a_radar.targets, targets);
		error_calculations(a_radar.targets, matching_target);
	}
	else
	{
		error_calculations(a_radar.targets, targets);
	}
}

void read_radar_signal(radar &a_radar, vector<valarray<complex<double>>> &y, istream& infile)
{
	int rows, cols;
	infile >> rows >> cols;
	y = vector<valarray<complex<double>>>(rows, valarray<complex<double>>(cols));
	for (int i = 0; i < rows; ++i)
	{
		for (int j = 0; j < cols; ++j)
		{
			double real, imag;
			infile >> real >> imag;
			y[i][j] = complex<double>(real, imag);
			// cout << y[i][j] << "\n";
		}
	}
}

int main(int argc, char** argv)
{
	if (argc != 2) {
		cerr << "Usage ./simulate [input filename]\n";
		return 0;
	}
	ifstream infile(argv[1]);
	if (!infile.is_open()) {
		cerr << "Input file cannot be found.\n";
		return 1;
	}
	vector<valarray<complex<double>>> y;
	radar a_radar;
	get_a_radar(a_radar, infile);
	read_radar_signal(a_radar, y, infile);
	infile.close();
	process_radar_signal(a_radar, y);
}

#endif /* SIMULATOR_CPP_ */
