#include "util.h"

inline double constrainAngle(double x)
{
	x = fmod(x + M_PI, 2 * M_PI);
	if (x < 0)
		x += 2 * M_PI;
	return x - M_PI;
}

inline double angleConv(double angle)
{
	return fmod(constrainAngle(angle), 2 * M_PI);
}

inline double angleDiff(double a, double b)
{
	double dif = fmod(b - a + M_PI, 2 * M_PI);
	if (dif < 0)
		dif += 2 * M_PI;
	return dif - M_PI;
}

inline double unwrap(double previousAngle, double newAngle)
{
	return previousAngle - angleDiff(newAngle, angleConv(previousAngle));
}

void unwrap(vector<double> &x)
{
	if (x.size() > 1)
	{
		for (size_t i = 1; i < x.size(); i++)
		{
			x[i] = unwrap(x[i - 1], x[i]);
		}
	}
}

vector<double> sampling(double min, double interval, double max)
{
	vector<double> samples;
	int N = floor((max - min) / interval);
	double t = 0.0;
	for (int i = 0; i <= N; i++)
	{
		t = min + interval * i;
		samples.push_back(t);
	}

	return samples;
}

void quadraticfit(vector<double> &dx, vector<double> &dy, vector<double> &coeffs)
{
	double S00, S01, S10, S11, S20, S21, S30, S40 = 0.0;
	S00 = S01 = S10 = S11 = S20 = S21 = S30 = S40;

	S00 = (dx.size());
	for (size_t i = 0; i < dx.size(); i++)
	{
		double dxi = dx[i];
		double dyi = dy[i];
		S10 += dxi;
		S20 += pow(dxi, 2);
		S30 += pow(dxi, 3);
		S40 += pow(dxi, 4);

		S01 += dyi;
		S11 += dxi * dyi;
		S21 += pow(dxi, 2) * dyi;
	}

	double a, b, c = 0.0;
	a = (S01 * S10 * S30 - S11 * S00 * S30 - S01 * pow(S20, 2) + S11 * S10 * S20 + S21 * S00 * S20 - S21 * pow(S10, 2)) / (S00 * S20 * S40 - pow(S10, 2) * S40 - S00 * pow(S30, 2) + 2 * S10 * S20 * S30 - pow(S20, 3));
	b = (S11 * S00 * S40 - S01 * S10 * S40 + S01 * S20 * S30 - S21 * S00 * S30 - S11 * pow(S20, 2) + S21 * S10 * S20) / (S00 * S20 * S40 - pow(S10, 2) * S40 - S00 * pow(S30, 2) + 2 * S10 * S20 * S30 - pow(S20, 3));
	c = (S01 * S20 * S40 - S11 * S10 * S40 - S01 * pow(S30, 2) + S11 * S20 * S30 + S21 * S10 * S30 - S21 * pow(S20, 2)) / (S00 * S20 * S40 - pow(S10, 2) * S40 - S00 * pow(S30, 2) + 2 * S10 * S20 * S30 - pow(S20, 3));

	coeffs.push_back(a);
	coeffs.push_back(b);
	coeffs.push_back(c);
}

double polyval(vector<double> q, double x)
{
	long double val = 0.0;
	int degree = q.size();
	for (int i = 0; i < degree; i++)
	{
		val += pow(x, degree - i - 1) * q[i];
	}

	return val;
}

double gaussian(double x, double mu, double sigma)
{
	return exp(-(((x - mu) / (sigma)) * ((x - mu) / (sigma))) / 2.0);
}

int nextpow2(double x)
{
	int p = 0;
	while (pow(2.0, p) <= x)
	{
		p = p + 1;
	}

	return p;
}

void initVector(int rows, int cols, vector<valarray<complex<double>>> &y)
{
	y.clear();
	for (int i = 0; i < rows; i++)
	{
		valarray<complex<double>> yi(cols);
		y.push_back(yi);
	}
}

void initVector(int N, vector<complex<double>> &y)
{
	y.clear();
	complex<double> zero(0.0, 0.0);
	for (int i = 0; i < N; i++)
	{
		y.push_back(zero);
	}
}

bool IsPowerofTwo(int x)
{
	return (x != 0) && ((x & (x - 1)) == 0);
}

double rand_uniform()
{
	return (double)rand() / (double)RAND_MAX;
}

void peakinterp(double &amp, double &del_index,
				double &z1, double &z2, double &z3)
{
	del_index = -0.5 * ((z3 - z1)) / (z1 - 2 * z2 + z3);
	int k = del_index;
	amp = 0.5 * ((k - 1) * k * z1 - 2 * (k - 1) * (k + 1) * z2 + (k + 1) * k * z3);
}

double db(complex<double> x, string label)
{
	double val = 0.0;
	double ab = abs(x);
	if (label == "voltage")
	{
		val = (ab == 0) ? 0 : 20 * log10(ab);
	}
	else if (label == "power")
	{
		val = (ab == 0) ? 0 : 10 * log10(ab);
	}
	else
	{
		cout << "invalid label of db" << endl;
	}

	return val;
}

vector<double> hamming(int L)
{
	vector<double> w;
	for (int i = 0; i < L; i++)
	{
		double val = 0.54 - 0.46 * cos(2.0 * M_PI * (double)i / (double)(L - 1));
		w.push_back(val);
	}
	return w;
}