/*
 * util.h
 *
 *  Created on: 22/05/2014
 *      Author: yan
 */

#ifndef UTIL_H_
#define UTIL_H_
#include "inc.h"

//Normalize to [-180,180):
extern inline double constrainAngle(double x);

// convert to [-360,360]
extern inline double angleConv(double angle);

extern inline double angleDiff(double a, double b);

extern inline double unwrap(double previousAngle, double newAngle);

extern void unwrap(vector<double> &x);

extern vector<double> sampling(double min, double interval, double max);

/* least square estimation to fit a quadratic function */
extern void quadraticfit(vector<double> &dx, vector<double> &dy, vector<double> &coeffs);

extern double polyval(vector<double> q, double x);

extern double gaussian(double x, double mu, double sigma);

extern int nextpow2(double x);

extern void initVector(int rows, int cols, vector<valarray<complex<double>>> &y);

extern void initVector(int N, vector<complex<double>> &y);

extern bool IsPowerofTwo(int x);

template <typename T>
void transpose(vector<valarray<T>> &x, vector<valarray<T>> &xt)
{

	int M = x.size();
	int N = x[0].size();
	initVector(N, M, xt);
	for (int i = 0; i < M; i++)
	{
		for (int j = 0; j < N; j++)
		{
			xt[j][i] = x[i][j];
		}
	}
}

/* fast Fourier transform for a vector
 * time domain -> frequency domain */
template <typename T>
void fft(valarray<T> &x, bool radix)
{
	const size_t N = x.size();

	if (radix)
	/* if signal size is the power of 2, use cooley-turkey*/
	{
		if (N <= 1)
			return;
		// divide
		valarray<T> even = x[std::slice(0, N / 2, 2)];
		valarray<T> odd = x[std::slice(1, N / 2, 2)];

		// conquer
		fft(even, radix);
		fft(odd, radix);

		// combine
		for (size_t k = 0; k < N / 2; ++k)
		{
			T t = std::polar(1.0, -2.0 * M_PI * (double)k / (double)N) * odd[k];
			x[k] = even[k] + t;
			x[k + N / 2] = even[k] - t;
		}
	}
	else
	/* slow ft */
	{
		valarray<T> out(N);
		for (size_t k = 0; k < N; k++)
		{
			T Xk(0.0, 0.0);
			for (size_t j = 0; j < N; j++)
			{
				double img = -2.0 * M_PI * (double)j * (double)k / (double)N;
				T v(0, img);
				Xk = Xk + x[j] * exp(v);
			}
			out[k] = Xk;
		}
		x = out;
	}
}

/* inverse Fourier transform
 * frequency domain -> time domain */
template <typename T>
void ifft(valarray<T> &x, bool radix)
{
	if (radix)
	/* if signal size is the power of 2, use cooley-turkey*/
	{
		// conjugate the complex numbers
		x = x.apply(std::conj);

		// forward fft
		fft(x, radix);

		// conjugate the complex numbers again
		x = x.apply(std::conj);

		// scale the numbers
		x /= (double)x.size();
	}
	else
	{
		/* slow ift */
		int N = x.size();
		valarray<T> out(N);
		for (int j = 0; j < N; j++)
		{
			T xj(0.0, 0.0);
			for (int k = 0; k < N; k++)
			{
				double img = 2.0 * M_PI * (double)j * (double)k / (double)N;
				T v(0, img);
				xj = xj + (1.0 / (double)N) * x[k] * exp(v);
			}
			out[j] = xj;
		}
		x = out;
	}
}

template <typename T>
void conv(valarray<T> &h, valarray<T> &y, valarray<T> &out)
{
	int M = y.size();
	int L = h.size();
	for (int i = 0; i < M + L; i++)
	{
		int i1 = i;
		for (int k = 0; k < L; k++)
		{
			if (i1 >= 0 && i1 < M)
				out[i] = out[i] + (y[i1] * h[k]);
			i1 = i1 - 1;
		}
	}
}

template <typename T>
void conv(valarray<T> &h,
		  vector<valarray<T>> &y,
		  vector<valarray<T>> &out, bool rows)
{
	int M = y.size();
	int N = y[0].size();
	int L = h.size();
	if (rows)
	/*convolution on rows*/
	{
		for (int i = 0; i < M; i++)
		{
			conv(h, y[i], out[i]);
		}
	}
	else
	/*convolution on columns*/
	{
		for (int j = 0; j < N; j++)
		{
			for (int i = 0; i < M + L; i++)
			{
				int i1 = i;
				for (int k = 0; k < L; k++)
				{
					if (i1 >= 0 && i1 < M)
						out[i][j] = out[i][j] + (y[i1][j] * h[k]);
					i1 = i1 - 1;
				}
			}
		}
	}
}

extern double rand_uniform();

template <typename T>
void git_rotate(vector<valarray<T>> &x, int num_places)
{
	int M = x.size();
	int N = x[0].size();

	if (N > 1)
	/* rotate columns */
	{
		num_places = num_places % N;
		for (int i = 0; i < M; i++)
		{
			for (int j = 0; j < N - num_places; j++)
			{
				T temp = x[i][j];
				x[i][j] = x[i][N - num_places + j];
				x[i][N - num_places + j] = temp;
			}
		}
	}
	/* rotate rows */
	else if (M > 1)
	{
		num_places = num_places % M;
		for (int i = 0; i < M - num_places; i++)
		{
			for (int j = 0; j < N; j++)
			{
				T temp = x[i][j];
				x[i][j] = x[M - num_places + i][j];
				x[M - num_places + i][j] = temp;
			}
		}
	}
}

/* peakinterp performs a quadratic interpolation of the peak
 * defined by three consecutive data values in the three-element vector z.
 * the middle value z2 must be the largest element.
 * peakinterp returns the interpolated peak amplitude and the interpolated
 * peak location relative to the center element in the range bins*/
extern void peakinterp(double &amp, double &del_index,
 				double &z1, double &z2, double &z3);

/* converts energy or power measurement to decibels */
extern double db(complex<double> x, string label);


/* L-point symmetric Hamming window in the column vector w*/
extern vector<double> hamming(int L);

#endif /* UTIL_H_ */
